#include <cstdlib>
#include <complex>
#include <conio.h>
#include <iostream>
#include "opti.hpp"
#include "pngwriter-0.5.4/src/pngwriter.h"

// Phased Gaussian convolution
// Parallel components image processing test and optimizer v1.1
//
// This work is placed in the public domain

using namespace std;
const double pi = 2*acos(0.0);

// Input image structure
struct ImageData {
  unsigned int 	 width;
  unsigned int 	 height;
  unsigned int 	 bytes_per_pixel; /* 3:RGB, 4:RGBA */
  unsigned char	 pixel_data[512 * 512 * 3 + 1];
};

extern ImageData lena; // Lena in lena.c -> liblena.a
extern ImageData hubble; // Hubble deep view in hubble.c -> libhubble.a

// Complex FFT and IFFT, with scaling in IFFT to enable convolution
void fft(int n, complex<double> *x, bool inv) {
    int s = (inv)? 1 : -1;
    int nxp, nxp2;
    for (nxp = n; (nxp2 = nxp/2) > 0; nxp = nxp2) {
        double wpwr = 2*pi/nxp;
        for (int m = 0; m < nxp2; ++m) {
            double argg = m * wpwr;
            complex<double> w(cos(argg), s*sin(argg));
            for (int jj1 = m; jj1+nxp-m <= n; jj1 += nxp) {
                int jj2 = jj1 + nxp2;
                complex<double> t = x[jj1] - x[jj2];
                x[jj1] += x[jj2];
                x[jj2] = t*w;
            }
        }
    }
    for (int i = 0, j = 0; i < n - 1; ++i) {
        if (i < j) {
            swap(x[i], x[j]);
        }
        int k;
        for (k = n/2; k <= j; k /= 2) {
            j -= k;
        }
        j += k;
    }
    if (inv) {
        for (int i = 0; i < n; ++i) {
            x[i] *= 1.0/n;
        }
    }
}

// Kernel parameter set
// The one-dimensional components are: (cos(x*x*c) * r + sin(x*x*c) * i) * exp(-b *x*x)
// where i is not the imaginary unit
class KernelParamSet {
private:
    int num;
    const double *params;
public:
    double b(int kernelIndex) const {
        return params[4*kernelIndex + 0];
    }
    double c(int kernelIndex) const {
        return params[4*kernelIndex + 1];
    }
    double r(int kernelIndex) const {
        return params[4*kernelIndex + 2];
    }
    double i(int kernelIndex) const {
        return params[4*kernelIndex + 3];
    }
    int numKernels() const {
        return num;
    }
    KernelParamSet(int num, const double *params) : num(num), params(params) {
    }
};

// Predefined equiripple flat disc parameter sets
const double disc1Params[] = {
    0.862325, 1.624835, 0.767583, 1.862321
};
const double disc2Params[] = {
    0.886528, 5.268909, 0.411259, -0.548794,
    1.960518, 1.558213, 0.513282, 4.561110
};
const double disc3Params[] = {
    2.176490, 5.043495, 1.621035, -2.105439,
    1.019306, 9.027613, -0.280860, -0.162882,
    2.815110, 1.597273, -0.366471, 10.300301
};
const double disc4Params[] = {
    4.338459, 1.553635, -5.767909, 46.164397,
    3.839993, 4.693183, 9.795391, -15.227561,
    2.791880, 8.178137, -3.048324, 0.302959,
    1.342190, 12.328289, 0.010001, 0.244650
};
const double disc5Params[] = {
    4.892608, 1.685979, -22.356787, 85.912460,
    4.711870, 4.998496, 35.918936, -28.875618,
    4.052795, 8.244168, -13.212253, -1.578428,
    2.929212, 11.900859, 0.507991, 1.816328,
    1.512961, 16.116382, 0.138051, -0.010000
};
const double disc6Params[] = { // This set could use extra digits
    5.143778, 2.079813, -82.326596, 111.231024,
    5.612426, 6.153387, 113.878661, 58.004879,
    5.982921, 9.802895, 39.479083, -162.028887,
    6.505167, 11.059237, -71.286026, 95.027069,
    3.869579, 14.810520, 1.405746, -3.704914,
    2.201904, 19.032909, -0.152784, -0.107988
};
const KernelParamSet disc1(1, disc1Params);
const KernelParamSet disc2(2, disc2Params);
const KernelParamSet disc3(3, disc3Params);
const KernelParamSet disc4(4, disc4Params);
const KernelParamSet disc5(5, disc5Params);
const KernelParamSet disc6(6, disc6Params);

// Freely play with this:
const double testParams[] = {
    4.892608, 1.685979, -22.356787, 85.912460,
    4.711870, 4.998496, 35.918936, -28.875618,
    4.052795, 8.244168, -13.212253, -1.578428,
    2.929212, 11.900859, 0.507991, 1.816328,
    1.512961, 16.116382, 0.138051, -0.010000
};
const KernelParamSet testSet(5, testParams);

// Optimization problem
class ParallelDiscProblem : public Opti::Problem {
private:
    int numKernels;
    double transitionBandwidth;
    double *min;
    double *max;
public:
    int getNumKernels() {
        return numKernels;
    }

    // Constructor
    ParallelDiscProblem(int numKernels, double transitionBandwidth) {
        this->transitionBandwidth = transitionBandwidth;
        this->numKernels = numKernels;
        min = new double[numKernels*4];
        max = new double[numKernels*4];
        for (int t = 0; t < numKernels; t++) {
            // Params: b, c, r, i
            min[t*4+0] = -10;
            max[t*4+0] = 10;

            min[t*4+1] = 0.1;
            max[t*4+1] = 100;

            min[t*4+2] = -15;
            max[t*4+2] = 15;

            min[t*4+3] = -15;
            max[t*4+3] = 15;
        }
    }

    // Destructor
    ~ParallelDiscProblem() {
        delete[] min;
        delete[] max;
    }

    // Return number of parameters to optimize
    //
    // Must be implemented in the actual optimization problem.
    int getNumDimensions() {
        return numKernels*4;
    }

    // The parameters being optimized are assumed to be within a range.
    // These functions give arrays containing minimum and maximum values
    // for the parameters. NOTE: Obtained solution may be outside these ranges,
    // but a solution will be found more easily if it is within these ranges.
    //
    // Must be implemented in the actual optimization problem.
    double *getMin() {
        return min;
    }
    double *getMax() {
        return max;
    }

    // "cost function" being minimized. Return cost for the set of parameters
    // given in the params array. The compare value is provided as an eariler
    // cost value the current cost value is compared to. If the current cost
    // evaluation is known to give a higher value than the compare value, the
    // evaluation can be interrupted prematurely to save computing time and
    // a value higher than or equal to the compare value should be returned.
    // The cost function is allowed to modify params to ensure parameter
    // constraints, wraparound, etc.
    //
    // Must be implemented in the actual optimization problem.
    virtual double costFunction(double *params, double compare) {
        // Params: b, c, r, i
        int numPassBandSamples = 200;
        int numStopBandSamples = 200;
        double maxerror = 0;
        for (int u = 0; u < numKernels; u++) {
            double b = params[u*4+0];
            if (b < 0.2) {
                b = 0.4 - b;
            }
            if (b > 20) {
                b = 40 - b;
            }
            params[u*4+0] = b;
            double c = params[u*4+1];
            if (c < 0.2) {
                c = 0.4 - c;
/*            } else if (c > 100) {
                c = 200 - c;*/
            }
            double centerFreq = (u*1.14 + 0.5) * pi;
            if (c < centerFreq-0.4*pi) {
              c = 2*(centerFreq-0.4*pi) - c;
            } else if (c > centerFreq+0.4*pi) {
              c = 2*(centerFreq+0.4*pi) - c;
            }
            params[u*4+1] = c;
            double r = params[u*4+2];
            if (fabs(r) < 0.01) {
                if (r > 0) {
                    r = 0.02-r;
                } else {
                    r = -0.02-r;
                }
            }
            params[u*4+2] = r;
            double i = params[u*4+3];
            if (fabs(i) < 0.01) {
                if (i > 0) {
                    i = 0.02-i;
                } else {
                    i = -0.02-i;
                }
            }
            params[u*4+3] = i;
        }
        for (int t = 0; t < numPassBandSamples; t++) {
            double output = 0;
            double x = t / (numPassBandSamples - 1.0);
            for (int u = 0; u < numKernels; u++) {
                double b = params[u*4+0];
                double c = params[u*4+1];
                double r = params[u*4+2];
                double i = params[u*4+3];
                output += (cos(x*x*c) * r + sin(x*x*c) * i) * exp(-b*x*x);
            }
            double error = fabs(output-1);
            if (maxerror < error) {
                maxerror = error;
                if (maxerror > compare) return maxerror;
            }
        }
        for (int t = 0; t < numStopBandSamples; t++) {
            double output = 0;
            double x = 1 + transitionBandwidth - log((numStopBandSamples - t) / (double)numStopBandSamples);
            for (int u = 0; u < numKernels; u++) {
                double b = params[u*4+0];
                double c = params[u*4+1];
                double r = params[u*4+2];
                double i = params[u*4+3];
                output += (cos(x*x*c) * r + sin(x*x*c) * i) * exp(-b*x*x);
            }
            double error = fabs(output);
            if (maxerror < error) {
                maxerror = error;
                if (maxerror > compare) return maxerror;
            }
        }
        return maxerror;
    }

    void print(double *params) {
        printf("Number of components: %d, transition bandwidth: %f, ripple: %f.\n", numKernels, transitionBandwidth, costFunction(params, 666));
        for (int t = 0; t < numKernels; t++) {
            double b = params[t*4+0];
            double c = params[t*4+1];
            double r = params[t*4+2];
            double i = params[t*4+3];
            printf("Component %d:\n(cos(x*x*%f) * %f + sin(x*x*%f) * %f) * exp(-%f*x*x)\n", t, c, r, c, i, b);
        }
        printf("\n");
        for (int t = 0; t < numKernels; t++) {
            double b = params[t*4+0];
            double c = params[t*4+1];
            double r = params[t*4+2];
            double i = params[t*4+3];
            printf("%f, %f, %f, %f\n", b, c, r, i);
        }
    }
};

int main(int argc, char *argv[])
{
    // Edit these to enable/disable optimizer and for optimization options
    const bool optimize = true; // Look for a parameter set rather than use a precalculated set, default: false
    const int optimizeNumKernels = 1; // How many kernels to have in the set to optimize, default: 2
    const float transitionBandwidth = 0.2; // Width of positive side transition band as a fraction of width of pass band positive side, defaul: 0.2

    // Optimize parallel phased Gaussian kernels for disc convolution
    double optimizedParams[optimizeNumKernels*4];
    if (optimize) {
        ParallelDiscProblem problem(optimizeNumKernels, transitionBandwidth);
        Opti::DE deOptimizer1(&problem, 1000, new Opti::MagnusRecombinator(3));
        Opti::DE deOptimizer2(&problem, 500, new Opti::DERecombinator(0.999, 0.76));
        Opti::G3 g3Optimizer(&problem, 200);
        Opti::GreedyMagnus greedyMagnusOptimizer(&problem, 200);

        // Edit this, choose optimization strategy from one of the above, default: deOptimizer2
        Opti::Strategy &optimizer(deOptimizer2);

        printf("\nOptimizing...\n\n");
        double bestCost;
        for(int t = 0; !kbhit(); t++) {
    	    bestCost = optimizer.evolve();
         	if (!(t % 1000)) {
    	      printf("generation=%d, bestCost=%.10f, average %.10f\n", t, bestCost, optimizer.averageCost());
            }
        }
        printf("Parameter vector printout:\n");
      	problem.print(optimizer.best());
      	for (int t = 0; t < optimizeNumKernels*4; t++) {
            optimizedParams[t] = optimizer.best()[t];
        }
    }
    KernelParamSet optimized(optimizeNumKernels, optimizedParams);

    // Edit these
    ImageData &inputImage(hubble); // Input image, Hubble deep view: hubble, Lena: lena
    KernelParamSet params(optimize ? optimized : disc5); // Parameterized kernel set, default: disc5
    const bool inputIsImpulse = true; // Use an impulse as input instead of an image, and normalize response, default: true
    const bool colorImpulse = false; // If input is impulse, is it colored or gray impulse, colored: true, gray: false
    const bool normalizeImage = true; // If input is an image, automatically increase brightness to maximum without clipping, default: false
    double scale = 32; // Spatial scaling of the kernels, default: 64
    const char outputFileName[] = "output.png"; // Output png file name, default: "output.png"

    // Initialize tables
    pngwriter png(512, 512, 0, outputFileName);
    double *imageInR = new double[512*512];
    double *imageInG = new double[512*512];
    double *imageInB = new double[512*512];
    double *imageOutR = new double[512*512];
    double *imageOutG = new double[512*512];
    double *imageOutB = new double[512*512];
    complex<double> *componentR = new complex<double>[512*512];
    complex<double> *componentG = new complex<double>[512*512];
    complex<double> *componentB = new complex<double>[512*512];
    complex<double> *kernels = new complex<double>[512*params.numKernels()];
  
    // Create kernels
    double accu = 0;
    for (int t = 0; t < params.numKernels(); t++) {
        double b = params.b(t);
        double c = params.c(t);
        for (int xInPixels = 0; xInPixels < 512; xInPixels++) {
            double x;
            if (xInPixels >= 256) {
                x = (xInPixels - 512)/scale;
            } else {
                x = xInPixels/scale;
            }
            complex<double> imagUnit(0, 1);
            kernels[t*512+xInPixels] = exp(-x*x*b + imagUnit*x*x*c);
        }
    }
    
    // Composite normalization (analysis)
    for (int y = 0; y < 512; y++) {
        for (int x = 0; x < 512; x++) {
            double temp2 = 0;
            for (int t = 0; t < params.numKernels(); t++) {
                double r = params.r(t);
                double i = params.i(t);
                complex<double> temp = kernels[t*512+x]*kernels[t*512+y];
                temp2 += r*real(temp) + i*imag(temp);
            }
            if (!inputIsImpulse) {
                accu += temp2;
            } else if (accu < fabs(temp2)) {
                accu = fabs(temp2);
            }
        }
    }

    // Composite normalization (scaling)
    for (int t = 0; t < params.numKernels(); t++) {
        for (int x = 0; x < 512; x++) {
            kernels[t*512+x] /= sqrt(accu);
        }
    }

    // Convert kernels to frequency domain
    for (int t = 0; t < params.numKernels(); t++) {
        fft(512, &kernels[t*512], false);
    }
    
    // Get input image
    if (inputIsImpulse) {
        // Impulse
        for (int y = 0; y < 512; y++) {
            for (int x = 0; x < 512; x++) {
                imageInR[y*512+x] = 0;
                imageInG[y*512+x] = 0;
                imageInB[y*512+x] = 0;
            }
        }
        if (colorImpulse) {
            imageInB[256*512+256] = 1;
            imageInG[256*512+256] = 1;
        } else {
            imageInR[256*512+256] = 1;
            imageInG[256*512+256] = 1;
            imageInB[256*512+256] = 1;
        }
    } else {
        // Input pixel data, linearize data with intensity (adjust for gamma).
        for (int y = 0; y < 512; y++) {
            for (int x = 0; x < 512; x++) {
                imageInR[y*512+x] = inputImage.pixel_data[(y*512+x)*3+0]/255.0;
                imageInR[y*512+x] *= imageInR[y*512+x];
                imageInG[y*512+x] = inputImage.pixel_data[(y*512+x)*3+1]/255.0;
                imageInG[y*512+x] *= imageInG[y*512+x];
                imageInB[y*512+x] = inputImage.pixel_data[(y*512+x)*3+2]/255.0;
                imageInB[y*512+x] *= imageInB[y*512+x];
                imageOutR[y*512+x] = 0;
                imageOutG[y*512+x] = 0;
                imageOutB[y*512+x] = 0;
            }
        }
    }

    // Clear output accumulator
    for (int y = 0; y < 512; y++) {
        for (int x = 0; x < 512; x++) {
            imageOutR[y*512+x] = 0;
            imageOutG[y*512+x] = 0;
            imageOutB[y*512+x] = 0;
        }
    }

    // Do the actual work: circularly symmetric convolution
    int lastPercent = -1;
    for (int t = 0; t < params.numKernels(); t++) {
        double r = params.r(t);
        double i = params.i(t);
            
        // Horizontal convolution
        for (int y = 0; y < 512; y++) {
            int percent = 100*(t*1024+y)/(params.numKernels()*1024);
            if (lastPercent != percent) {
                lastPercent = percent;
                printf("\r%d%%", percent);
            }
            for (int x = 0; x < 512; x++) {
                componentR[y*512+x] = imageInR[y*512+x]; // Copy from image for convolution
                componentG[y*512+x] = imageInG[y*512+x];
                componentB[y*512+x] = imageInB[y*512+x]; 
            }
            fft(512, &componentR[y*512], false); // Convert pixels to frequency domain
            fft(512, &componentG[y*512], false);
            fft(512, &componentB[y*512], false);
            for (int x = 0; x < 512; x++) {
                componentR[y*512+x] *= kernels[t*512+x]; // Multiply in frequency domain = convolve pixels
                componentG[y*512+x] *= kernels[t*512+x]; 
                componentB[y*512+x] *= kernels[t*512+x]; 
            }
            fft(512, &componentR[y*512], true); // Convert frequency domain to pixels
            fft(512, &componentG[y*512], true);
            fft(512, &componentB[y*512], true);
        }
        
        // Swap x and y
        for (int y = 1; y < 512; y++) {
            for (int x = 0; x < y; x++) {
                swap(componentR[y*512+x], componentR[x*512+y]);
                swap(componentG[y*512+x], componentG[x*512+y]);
                swap(componentB[y*512+x], componentB[x*512+y]);
            }
        }

        // Vertical convolution (done horizontally because x and y have been swapped)
        for (int y = 0; y < 512; y++) {
            int percent = 100*(t*1024+512+y)/(params.numKernels()*1024);
            if (lastPercent != percent) {
                lastPercent = percent;
                printf("\r%d%%", percent);
            }
            fft(512, &componentR[y*512], false); // Convert pixels to frequency domain
            fft(512, &componentG[y*512], false);
            fft(512, &componentB[y*512], false);
            for (int x = 0; x < 512; x++) {
                componentR[y*512+x] *= kernels[t*512+x]; // Multiply in frequency domain = convolve pixels
                componentG[y*512+x] *= kernels[t*512+x]; 
                componentB[y*512+x] *= kernels[t*512+x]; 
            }
            fft(512, &componentR[y*512], true); // Convert frequency domain to pixels
            fft(512, &componentG[y*512], true);
            fft(512, &componentB[y*512], true);
        }

        // Swap x and y, back to normal
        for (int y = 1; y < 512; y++) {
            for (int x = 0; x < y; x++) {
                swap(componentR[y*512+x], componentR[x*512+y]);
                swap(componentG[y*512+x], componentG[x*512+y]);
                swap(componentB[y*512+x], componentB[x*512+y]);
            }
        }
        
        // Accumulate components into composite
        for (int y = 0; y < 512; y++) {
            for (int x = 0; x < 512; x++) {
                imageOutR[y*512+x] += real(componentR[y*512+x])*r + imag(componentR[y*512+x])*i;
                imageOutG[y*512+x] += real(componentG[y*512+x])*r + imag(componentG[y*512+x])*i;
                imageOutB[y*512+x] += real(componentB[y*512+x])*r + imag(componentB[y*512+x])*i;
            }
        }
    }
    printf("\r100%%");

    accu = 0;
    if (!inputIsImpulse && normalizeImage) {
        for (int y = 0; y < 512; y++) {
            for (int x = 0; x < 512; x++) {
                if (imageOutR[y*512+x] > accu) accu = imageOutR[y*512+x];
                if (imageOutG[y*512+x] > accu) accu = imageOutG[y*512+x];
                if (imageOutB[y*512+x] > accu) accu = imageOutB[y*512+x];
            }
        }
    }
    if (accu == 0) {
        accu = 1;
    }

    // Copy output pixels to PNG
    for (int y = 0; y < 512; y++) {
        for (int x = 0; x < 512; x++) {
            double r = 0, g = 0, b = 0;
            if (!inputIsImpulse){
                r = imageOutR[y*512+x] / accu;
                g = imageOutG[y*512+x] / accu;
                b = imageOutB[y*512+x] / accu;
            } else if (colorImpulse) {
                // Mess with colors to show negative and over-unity values in a meaningful way
                if (imageOutR[y*512+x] < 0) {
                    g += -0.5*imageOutR[y*512+x];
                    b += -0.5*imageOutR[y*512+x];
                } else if (imageOutR[y*512+x] > 1) {
                    r += imageOutR[y*512+x];
                    g += 0.5*(imageOutR[y*512+x]-1);
                    b += 0.5*(imageOutR[y*512+x]-1);
                } else {
                    r += imageOutR[y*512+x];
                }
                if (imageOutG[y*512+x] < 0) {
                    r += -0.5*imageOutG[y*512+x];
                    b += -0.5*imageOutG[y*512+x];
                } else if (imageOutG[y*512+x] > 1) {
                    g += imageOutG[y*512+x];
                    r += 0.5*(imageOutG[y*512+x]-1);
                    b += 0.5*(imageOutG[y*512+x]-1);
                } else {
                    g += imageOutG[y*512+x];
                }
                if (imageOutB[y*512+x] < 0) {
                    r += -0.5*imageOutB[y*512+x];
                    g += -0.5*imageOutB[y*512+x];
                } else if (imageOutB[y*512+x] > 1) {
                    b += imageOutB[y*512+x];
                    r += 0.5*(imageOutB[y*512+x]-1);
                    g += 0.5*(imageOutB[y*512+x]-1);
                } else {
                    b += imageOutB[y*512+x];
                }
            } else {
                r = imageOutR[y*512+x]*0.5+0.5;
                g = imageOutG[y*512+x]*0.5+0.5;
                b = imageOutB[y*512+x]*0.5+0.5;
            }
            png.plot(x+1, 512-y, sqrt(r), sqrt(g), sqrt(b));
        }
    }

    // Bye bye
    png.close();
    system("PAUSE");
    return EXIT_SUCCESS;
}
